//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* expr.h: expressions                                                    *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://apps.jcns.fz-juelich.de/frida                                   *//
//**************************************************************************//

#include <boost/shared_ptr.hpp>

#define PTree boost::shared_ptr<class CTree>
#define PRef  boost::shared_ptr<class CRef>
#define PRgr  boost::shared_ptr<class CRegRef>


//! A Context specifies how a tree shall be evaluated.

class CContext {
 public:
    uint k, j, i;                   // file,scan,point we are refering to
    const vector<double> *vt;       // a vector we may refer to
    double vt_step;                 // step of equidistant vt, otherwise 0
    enum TDim { _1, _VI, _VT } dim; // kind of requested output
    bool want_error;                // errors requested?
    uint nv;                        // number of requested vector elements

    CContext( uint _k=-1, uint _j=-1, uint _i=-1 );

    void request_VT( const vector<double> *_vt );
    void request_VI( uint _nv );

    string context_info() const;
};


//! Return value for tree evaluation.

class CResult {
 public:
    bool vectorial;
    double r;
    double dr;
    vector<double> v;
    vector<double> dv;

    CResult() : vectorial(false), dr(0) {};

    void set_r( double _r ) {
        vectorial = false; r = _r; };
    void preset_v( int n ) {
        vectorial = true; v.resize(n); };
    void set_v( const vector<double>& _v ) {
        vectorial = true; v = _v; };

    double to_r( int i ) const {
        return vectorial ? v[i] : r; };
    double to_dr( int i ) const {
        return vectorial ? ( dv.size() ? dv[i] : 0 ) : dr; };
    bool has_err() const {
        return vectorial ? dv.size() : dr; };

    string result_info() const;
};


//! References to data points, indices, ...; used as endpoints of CTree's.

class CRef {
 public:
    PTree tk;
    PTree tj;
    PTree ti;

    CRef( PTree _tk=PTree(), PTree _tj=PTree(), PTree _ti=PTree() ) 
        : tk(_tk), tj(_tj), ti(_ti) {};
    
    uint get_k( const CContext& ctx ) const;
    uint get_j( const CContext& ctx, uint nj ) const;
    string ref_info() const;
};


//! Parsed generic expression.

class CTree {
 public:
    virtual void npar_exec( uint *np ) const =0;
    //! Recursive: set *kd=true if a subnode is K dependent
    virtual void kdep_exec( bool *kd ) const {;} // default: do nothing

    CTree() {};

    uint npar() const;
    bool k_dependent() const;
    virtual bool has_dummy() const =0; // does node depend on t ?

    //! Evaluate tree. Private. User calls pass through API implemented below.
    virtual void tree_val( CResult& ret, const CContext& ctx ) const =0;
    //! API for tree evaluation: compute one index.
    void tree_uival( uint *ret, const CContext& ctx ) const;
    //! API for tree evaluation: compute one numeric value.
    void tree_point_val(
        double *ret, double *dret,
        uint k=(uint)-1, uint j=(uint)-1, uint i=(uint)-1 ) const;
    //! API for tree evaluation: compute an array.
    void tree_vec_val(
        vector<double> *ret, vector<double> *dret,
        uint k=(uint)-1, uint j=(uint)-1 ) const;

    virtual void set_coord( CCoord& ret, uint k=(uint)-1 ) const =0;
    virtual string tree_info() const =0; // introspection, for debugging
};
