//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* node.h: subclasses for expression                                      *//
//* (C) Joachim Wuttke 2012-                                               *//
//* http://apps.jcns.fz-juelich.de/frida                                   *//
//**************************************************************************//


//! Function/operator node.

class CNodeFun: public CTree {
 private:
    static const uint maxarg=3;
    const class CFunc* fun;
    uint narg;
    PTree arg[maxarg];
    void npar_exec( uint *np ) const;
 public:
    CNodeFun( const class CFunc *_fun, PTree a0 );
    CNodeFun( const class CFunc *_fun, PTree a0, PTree a1 );
    CNodeFun( const class CFunc *_fun, PTree a0, PTree a1, PTree a2 );
    void tree_val( CResult& ret, const CContext& ctx ) const;
    bool has_dummy() const;
    void kdep_exec( bool *kd ) const;
    void set_coord( CCoord& ret, uint k ) const;
    string tree_info() const;
};


//! End node with fixed numeric value.

class CNodeVal: public CTree {
 private:
    double val;
    void npar_exec( uint *np ) const {;}
 public:
    CNodeVal( double _val );
    void tree_val( CResult& ret, const CContext& ctx ) const {
        ret.set_r( val ); }
    bool has_dummy() const { return false; }
    void set_coord( CCoord& ret, uint k ) const;
    string tree_info() const;
};


//! Indexed variable node.

class CNodeIva: public CTree {
 private:
    PVar var;
    PRef ref;
    void npar_exec( uint *np ) const;
 public:
    CNodeIva( PVar _var, PRef _ref=PRef( new CRef() ) ) 
        : var(_var), ref(_ref) {};
    void tree_val( CResult& ret, const CContext& ctx ) const;
    bool has_dummy() const { return false; }
    void kdep_exec( bool *kd ) const { *kd = true; }
    void set_coord( CCoord& ret, uint k ) const;
    string tree_info() const { return "ref[" + var->var_info() + "]"; }
};


//! Register reference node.

class CNodeRgr: public CTree {
 private:
    PRgr rgr;
    void npar_exec( uint *np ) const {;}
 public:
    CNodeRgr( PRgr& _rgr );
    void tree_val( CResult& ret, const CContext& ctx ) const;
    bool has_dummy() const { return false; }
    void set_coord( CCoord& ret, uint k ) const;
    string tree_info() const { return "$"; }
};


//! Function argument (t) node.

class CNodeDummy: public CTree {
 private:
    void npar_exec( uint *np ) const {;}
 public:
    CNodeDummy() {};
    void tree_val( CResult& ret, const CContext& ctx ) const {
        if ( !(ctx.vt) ) throw "BUG: tree_val with *vt=0";
        ret.set_v( *(ctx.vt) ); }
    bool has_dummy() const { return true; } // this node _is_ t, so it has t
    void set_coord( CCoord& ret, uint k ) const {
        throw "BUG: CNodeDummy::set_coord not expected to be needed"; }
    string tree_info() const { return "farg"; }
};


//! Curve evaluation node.

class CNodeCev: public CTree {
 private:
    PRef ref;
    PTree arg;
    void npar_exec( uint *np ) const {;}
 public:
    CNodeCev( PRef& _ref, PTree _arg );
    void tree_val( CResult& ret, const CContext& ctx ) const;
    bool has_dummy() const { return arg->has_dummy(); }
    void set_coord( CCoord& ret, uint k ) const;
    string tree_info() const { return "c[" + ref->ref_info() + "]"; }
};


//! Convolution with one arguments: shift.

class CNodeConv1: public CTree {
protected:
    PTree shift;
 public:
    CNodeConv1( const PTree& _shift ) : shift(_shift) {}
    void tree_val( CResult& ret, const CContext& ctx ) const;
    virtual void copy_theory( CResult& ret, const CContext& ctx,
                              double theshift ) const =0;
    virtual void convolve( CResult& ret, const CContext& ctx, double theshift,
                           const PSpec& sv, double conv_norm,
                           double conv_step ) const =0;
};

//! Convolution with two arguments: theory and shift.

class CNodeConv2: public CNodeConv1 {
protected:
    PTree theory;
    void npar_exec( uint *np ) const;
 public:
    CNodeConv2( const PTree& _theory, const PTree& _shift )
        : CNodeConv1( _shift ), theory(_theory) {}
    bool has_dummy() const { return theory->has_dummy(); }
    void kdep_exec( bool *kd ) const;
    void set_coord( CCoord& ret, uint k ) const { theory->set_coord( ret, k ); }
};

//! Convoluted function node.

class CNodeConv: public CNodeConv2 {
 public:
    CNodeConv( const PTree& _theory,
               const PTree& _shift=PTree( new CNodeVal( 0.0 ) ) )
        : CNodeConv2( _theory, _shift ) {}
    void copy_theory( CResult& ret, const CContext& ctx,
                      double theshift ) const;
    void convolve( CResult& ret, const CContext& ctx, double theshift,
                   const PSpec& sv, double conv_norm, double conv_step ) const;
    string tree_info() const {
        return "conv(" + theory->tree_info() +","+ shift->tree_info() + ")"; }
};


//! Convolution with primitive.

class CNodePConv: public CNodeConv2 {
 public:
    CNodePConv( const PTree& _theory,
                const PTree& _shift=PTree( new CNodeVal( 0.0 ) ) )
        : CNodeConv2( _theory, _shift ) {};
    void copy_theory( CResult& ret, const CContext& ctx,
                      double theshift ) const;
    void convolve( CResult& ret, const CContext& ctx, double theshift,
                   const PSpec& sv, double conv_norm, double conv_step ) const;
    string tree_info() const {
        return "pconv(" + theory->tree_info() +","+ shift->tree_info() + ")"; }
};


//! Dirac's delta function node, to request a copy of the resolution function.

class CNodeDirac: public CNodeConv1 {
 private:
    void npar_exec( uint *np ) const;
 public:
    CNodeDirac( const PTree& _shift=PTree( new CNodeVal( 0.0 ) ) )
        : CNodeConv1( _shift ) {};
    void copy_theory( CResult& ret, const CContext& ctx,
                      double theshift ) const;
    void convolve( CResult& ret, const CContext& ctx, double theshift,
                   const PSpec& sv, double conv_norm, double conv_step ) const;
    bool has_dummy() const { return true; } // has implicit t dependence
    void kdep_exec( bool *kd ) const;
    void set_coord( CCoord& ret, uint k ) const;
    string tree_info() const {
        return "resol(" + shift->tree_info() + ")"; }
};
